const EVENT_COLOR = "#LightYellow";
const HTTP_COLOR = "#Peru";
const SCRIPT_COLOR = "#Olive;text:white";
const BOT_COLOR = "#Blue;text:white";
const EMAIL_SENDER_COLOR = "#Red;text:white";

const EVENT_NAME = ": Event";
const HTTP_NAME = ": Http";
const SCRIPT_NAME = ": Script";
const BOT_NAME = ": Chat";
const EMAIL_SENDER_NAME = ": EmailSender";

function isHttp(actions) {
  return actions
    .map((action) => action.actionType)
    .some((type) => type == "HttpRequest");
}

function isScript(actions) {
  return actions
    .map((action) => action.actionType)
    .some((type) => type == "Script");
}

function isChat(actions) {
  return actions
      .map((action) => action.actionType)
      .some((type) => type == "Chat");
}

function isEmailSender(actions) {
  return actions
      .map((action) => action.actionType)
      .some((type) => type == "EmailSender");
}

function isEmpty(arr) {
  return arr === undefined || arr == null || arr.length == 0;
}

function isNotBlank(str) {
  return str !== undefined && str !== null && str.trim() !== "";
}

function drawState(eventState, state) {
  let sb = "";
  const name = state.name;
  const actions = state.actions;
  const handlers = state.eventHandlers;
  if (isEmpty(actions)) {
    if (isEmpty(handlers)) {
      throw Error(`state ${name} handlers must not empty`);
    }
    let events = new Set()
    for (const handle of handlers) {
      const stateName = handle.eventType;
      events.add(stateName)
      sb += `state ${stateName} ${EVENT_COLOR} ${EVENT_NAME}\n`;
    }
    eventState.set(name, events);
  } else {
    if (isHttp(actions)) {
      sb += `state ${name} ${HTTP_COLOR} ${HTTP_NAME}\n`;
    } else if (isScript(actions)) {
      sb += `state ${name} ${SCRIPT_COLOR} ${SCRIPT_NAME}\n`;
    } else if (isChat(actions)) {
      sb += `state ${name} ${BOT_COLOR} ${BOT_NAME}\n`;
    } else if (isEmailSender(actions)) {
      sb += `state ${name} ${EMAIL_SENDER_COLOR} ${EMAIL_SENDER_NAME}\n`;
    } else {
      throw Error(`state ${name} actions is incorrect`);
    }
  }
  return sb;
}

function append(from, to, description) {
  if (isNotBlank(description)) {
    description = ": " + description;
  }
  return `${from} ---> ${to}${description}\n`;
}

function wrapCond(cond) {
  if (isNotBlank(cond)) {
    return "(" + cond + ")";
  } else {
    return null;
  }
}

function drawTransitionCondition(eventCondition, transition) {
  const list = [];
  if (eventCondition !== undefined && eventCondition !== null) {
    list.push(eventCondition);
  }
  const transitionName = transition.transitionName;
  if (transitionName !== undefined && transitionName !== null) {
    list.push(`(TransitionName == ${transitionName})`);
  }
  const condition = transition.condition;
  if (condition !== undefined && condition !== null) {
    list.push(condition);
  }
  return list.join(" && ");
}

function drawExitCondition(eventCondition, exit) {
  const list = [];
  if (eventCondition !== undefined && eventCondition !== null) {
    list.push(eventCondition);
  }
  if (exit.condition !== undefined && exit.condition !== null) {
    list.push(exit.condition);
  }
  return list.join(" && ");
}

function drawStateLine(eventState, state) {
  const handlers = state.eventHandlers;
  if (isEmpty(handlers)) {
    return;
  }

  let sb = "";
  const name = state.name;
  for (const handler of handlers) {
    let sourceState;
    if (eventState.has(name)) {
      sourceState = handler.eventType;
    } else {
      sourceState = name;
    }
    const eventCondition = wrapCond(handler.condition);
    const actions = handler.actions;
    for (const action of actions) {
      const actionType = action.actionType
      let targetState = action.targetState
      if (actionType === "Transition") {
        if (eventState.has(targetState)) {
          let events = eventState.get(targetState)
          if(events.size > 1) {
            continue
          } else {
            targetState = events.values().next().value
          }
        }
        const desc = drawTransitionCondition(eventCondition, action);
        sb += append(sourceState, targetState, desc);
      } else if (actionType === "Exit") {
        const desc = drawExitCondition(eventCondition, action);
        sb += append(sourceState, "[*]", desc);
      }
    }
  }
  return sb;
}

function transformDSL(definition) {
  let sb = "";
  const eventState = new Map();
  const states = definition["states"];
  for (const state of states) {
    sb += drawState(eventState, state);
  }
  for (const state of states) {
    sb += drawStateLine(eventState, state);
  }
  return `@startuml

${sb}
@enduml
      `;
}

module.exports = { transformDSL };
